{-
 - Author: Dobromir Stefanov
 - Matric. number: 0834826
 -
 - Description:
 -
 - I initially started this as a ray tracer (and it still is one), but due to a
 - lack of time I couldn't be very creative with the scene that's being traced.
 - Therefore, I reimplemented a simple/minimal version of turtle graphics for
 - 3D space (basically rewriting the code from Tutorial 7) and had it render a
 - Sierpinski Triangle using a series of spheres instead of line segments. The
 - spheres are still ray traced, and that (combined with lack of optimisation)
 - is why the thing is so slow. (On 1000x1000 it renders step 5 of the L-system
 - Sierpinski Triangle definition in about 4 minutes on a fairly fast machine.
 -
 - The ray tracer itself supports more than the Sierpinski Triangle shows. It
 - supports reflection, full-screen anti-aliasing, and shadows, which are all
 - either disabled in this rendering for performance reasons, or simply not
 - visible. You can see these features by switching to the commented version of
 - the main function, increasing reflectionLevel (to 2 for instance), and
 - increasing anti-aliasing (note that the latter exponentially slows down
 - rendering time - setting it over 4 is not advised).
 -
 - It can obviously also render any sequence of turtle commands or L-system.
 - It should be noted, however, that it does not auto-scale the drawing as
 - the tutorial one does. If you change the rendered commands, it will probably
 - go partially or fully off-screen. It is theoretically able to render 3D
 - L-systems, though I have not tested it.
 -
 - In short, my entry is perhaps more interesting for the way it works than
 - what it produces as output.
 -
 - While I didn't manage to even come close to "finishing" it, it has been a
 - pleasure working on this. I extend thanks to my fellow students who provided
 - advice and help, and similarly to the ones that provided.. amusement :)
 - I would especially like to thank Prof. Wadler for teaching us such a
 - wonderful and inspiring language!
 -
 - Usage:
 -
 - There is a run.sh script provided that compiles and runs it. Alternatively,
 - compile with:
 -
 -     ghc -O -threaded -package parallel
 -
 - and run with arguments:
 -
 -     +RTS -N2 > file.ppm
 -
 - where 2 is the number of cores/CPUs the machine has, and file.ppm is a
 - suitable file to write to. (The script outputs to "output.ppm") If the file
 - is ommitted, it will write a PPM file to standard output.
 -
 - The output is in PPM format, which graphics programs like Photoshop and The
 - GIMP can read and convert. The "convert" utility from ImageMagick can also
 - convert it to e.g. a PNG ("convert output.ppm output.png")
 -
 - Output in 3000x3000 which took over 30 minutes to render can be seen here:
 - 
 -     http://img368.imageshack.us/img368/29/sierp2ou5.png
 -
 - A really high-res render would be interesting as you would be able to zoom
 - in and see the actual spheres, but this isn't very practical with this
 - unoptimised implementation.
 -
 - Immediately below are a few tweakable options, most notably the resolution.
 -
 -}

import Control.Parallel.Strategies

----------------------------
-- Some rendering options --
----------------------------

widthI, heightI :: Int
widthI = 1000
heightI = 1000

backgroundColor :: Color
backgroundColor = Color 0 0 0

shadowColor :: Color
shadowColor = Color 0 0 0

ambientLight :: Color
ambientLight = Color 0.1 0.1 0.1

reflectionLevel :: ReflectionLevel
reflectionLevel = 0

antiAliasing :: Int
antiAliasing = 1

----

viewer :: Point
viewer = (Vector 0 0 0)

width, height :: Double
width = fromIntegral widthI
height = fromIntegral heightI

aspectRatio :: Double
aspectRatio = width / height

pixelWidth, pixelHeight :: Distance
pixelWidth = imagePlaneWidth / width
pixelHeight = imagePlaneHeight / height

imagePlaneWidth, imagePlaneHeight :: Distance
imagePlaneHeight = 2
imagePlaneWidth = imagePlaneHeight * aspectRatio

fieldOfView :: Angle
fieldOfView = 60 * pi / 180

focalDistance :: Distance
focalDistance = (imagePlaneWidth / 2) / tan (fieldOfView / 2)

-----------------------
-- General functions --
-----------------------

doublify :: (a -> a -> b) -> (a -> b)
doublify f a = f a a

none :: (a -> Bool) -> [a] -> Bool
none f = not . any f

just :: Maybe a -> a
just (Just x) = x
just Nothing = error "error in 'just'"

first :: (a -> Bool) -> [a] -> [a]
first _ [] = []
first f (x:xs) | f x		= [x]
               | otherwise	= first f xs

allExcept :: Eq a => a -> [a] -> [a]
allExcept a = filter (/= a)

clamp :: Ord a => a -> a -> a -> a
clamp low high n = max low $ min high n

-----------------------------------------
-- Custom type definitions - raytracer --
-----------------------------------------

-- Vector

data Vector = Vector {	x :: Double,
			y :: Double,
			z :: Double
			} deriving (Eq, Show)

vmap :: (Double -> Double) -> Vector -> Vector
vmap f (Vector x y z) = Vector (f x) (f y) (f z)

v2map :: (Double -> Double -> Double) -> Vector -> Vector -> Vector
v2map f (Vector x1 y1 z1) (Vector x2 y2 z2) = Vector (f x1 x2) (f y1 y2) (f z1 z2)

vsmap :: (Double -> Double -> Double) -> Vector -> Double -> Vector
vsmap f (Vector x y z) n = v2map f (Vector x y z) (Vector n n n)

vuminus :: Vector -> Vector
vuminus = vmap negate

vplus, vminus :: Vector -> Vector -> Vector
vplus = v2map (+)
vminus = v2map (-)

vsplus, vsminus, vsmultiply, vsdivide :: Vector -> Double -> Vector
vsplus = vsmap (+)
vsminus = vsmap (-)
vsmultiply = vsmap (*)
vsdivide = vsmap (/)

vdot :: Vector -> Vector -> Double
vdot (Vector x1 y1 z1) (Vector x2 y2 z2) = x1*x2 + y1*y2 + z1*z2

vlength :: Vector -> Double
vlength = sqrt . doublify vdot

vcross :: Vector -> Vector -> Vector
vcross (Vector x1 y1 z1) (Vector x2 y2 z2) = Vector	(y1*z2 - z1*y2)
							(z1*x2 - x1*z2)
							(x1*y2 - y1*x2)

vnormalize :: Vector -> Vector
vnormalize x = vmap (/ (vlength x)) x

-- Point

type Point = Vector -- Points have all properties that vectors do

-- Color

data Color = Color { r :: Double, g :: Double, b :: Double } deriving (Eq, Show)

cmap :: (Double -> Double) -> Color -> Color

cmap f (Color x y z) = Color (f x) (f y) (f z)

c2map :: (Double -> Double -> Double) -> Color -> Color -> Color
c2map f (Color x1 y1 z1) (Color x2 y2 z2) = Color (f x1 x2) (f y1 y2) (f z1 z2)

csmap :: (Double -> Double -> Double) -> Color -> Double -> Color
csmap f (Color x y z) n = c2map f (Color x y z) (Color n n n)

cuminus :: Color -> Color
cuminus = cmap negate

cplus, cminus :: Color -> Color -> Color
cplus = c2map (+)
cminus = c2map (-)
cmultiply = c2map (*)

csplus, csminus, csmultiply, csdivide :: Color -> Double -> Color
csplus = csmap (+)
csminus = csmap (-)
csmultiply = csmap (*)
csdivide = csmap (/)

csum :: [Color] -> Color
csum = foldr1 (cplus)

-- Ray

data Ray = Ray { origin :: Point, direction :: Point } deriving (Eq, Show)

-- Material

data Material = Material { ambientReflectiveness :: Double, diffuseReflectiveness :: Double, specularReflectiveness :: Double, shininess :: Double, reflectiveness :: Double, transparency :: Double, indexOfRefraction :: Double } deriving (Eq, Show)

-- Object

data Triangle = Triangle { p1 :: Point, p2 :: Point, p3 :: Point } deriving (Eq, Show)
data Sphere = Sphere { center :: Point, radius :: Distance } deriving (Eq, Show)

data Object' = Triangle' Triangle | Sphere' Sphere deriving (Eq, Show)
data Object = Object {oObject :: Object', oMaterial :: Material, oColor :: Color} deriving (Eq, Show)

oNormal :: Object -> Point -> Vector
oNormal o p = oNormal' (oObject o) p
	where	oNormal' :: Object' -> Point -> Vector
		oNormal' (Sphere' s) p = (center s) `vminus` p
		oNormal' (Triangle' t) _ = ((p2 t) `vminus` (p1 t)) `vcross` ((p3 t) `vminus` (p1 t))

-- Light

data Light = Light { position :: Point, diffuse :: Color, specular :: Color } deriving (Eq, Show)

-- Type redefinitions for clarity

type Distance = Double
type Angle = Double
type ReflectionLevel = Int


----------------------------
-- Raytracing definitions --
----------------------------

primaryRayThroughPixel :: Int -> Int -> Ray
primaryRayThroughPixel x y
	| 0 <= x && x < widthI && 0 <= y && y < heightI
		= Ray (viewer) (vnormalize $ Vector
			(fromIntegral x * pixelWidth - imagePlaneWidth / 2 + pixelWidth / 2)
			(fromIntegral (heightI - y) * pixelHeight - imagePlaneHeight / 2 + pixelHeight / 2)
			(focalDistance))
	| otherwise = error $ "primaryRayThroughPixel called on (" ++ show x ++ ", " ++ show y ++ ") outside of range"

primaryRayThroughPixelAA :: Int -> Int -> Int -> Int -> Int -> Ray
primaryRayThroughPixelAA x y aaCycleX aaCycleY aaTotal
	| 0 <= x && x < widthI && 0 <= y && y < heightI && 0 < aaCycleX && 0 < aaCycleX && aaCycleX <= aaTotal && aaCycleY <= aaTotal
		= Ray (viewer) (vnormalize $ Vector
			(fromIntegral x * pixelWidth - imagePlaneWidth / 2 + pixelWidth * (fromIntegral aaCycleX / fromIntegral aaTotal))
			(fromIntegral (heightI - y) * pixelHeight - imagePlaneHeight / 2 + pixelHeight * (fromIntegral aaCycleY / fromIntegral aaTotal))
			(focalDistance))
	| otherwise = error $ "primaryRayThroughPixel called on (" ++ show x ++ ", " ++ show y ++ ") outside of range"

intersectRayTriangle :: Ray -> Triangle -> Maybe (Point, Point)
intersectRayTriangle (Ray origin direction) (Triangle p1 p2 p3)
	| isPointInsideTriangle intersectionPoint	= Just (just intersectionPoint, just intersectionPoint)
	| otherwise					= Nothing
	where	just :: Maybe a -> a
		just (Just x) = x
		normal :: Vector
		normal = vnormalize $ (p2 `vminus` p1) `vcross` (p3 `vminus` p1)
		intersectionPoint :: Maybe Point
		intersectionPoint
			| n_dot_d == 0 || t < 0	= Nothing
			| otherwise		= Just ((direction `vsmultiply` t) `vplus` origin)
			where	n_dot_d = normal `vdot` direction
				d = normal `vdot` p1
				t = (d - (normal `vdot` origin)) / n_dot_d
		isPointInsideTriangle :: Maybe Point -> Bool
		isPointInsideTriangle Nothing = False
		isPointInsideTriangle (Just p)
			|    normal `vdot` v1 < 0
			  || normal `vdot` v2 < 0
			  || normal `vdot` v3 < 0
							= False
			| otherwise			= True
			where	v1 = (p2 `vminus` p1) `vcross` (p `vminus` p1)
				v2 = (p3 `vminus` p2) `vcross` (p `vminus` p2)
				v3 = (p1 `vminus` p3) `vcross` (p `vminus` p3)

intersectRaySphere :: Ray -> Sphere -> Maybe (Point, Point)
intersectRaySphere (Ray origin direction) (Sphere center radius)
	| d < 0 || t0 < 0	= Nothing -- Does not intersect
	| t0 == t1		= Just (p0, p0) -- Tangent
	| otherwise		= Just (p0, p1) -- Intersects twice
	where	o = origin `vminus` center
		a = direction `vdot` direction
		b = 2 * (o `vdot` direction)
		c = o `vdot` o - radius^2
		d = b^2 - 4*a*c
		q | b < 0	= -0.5 * (b - sqrt d)
		  | otherwise	= -0.5 * (b + sqrt d)
		t0 = q/a
		t1 = c/q
		p0' = origin `vplus` direction `vsmultiply` t0
		p1' = origin `vplus` direction `vsmultiply` t1
		p0'' = vlength $ p0' `vminus` origin
		p1'' = vlength $ p1' `vminus` origin
		p0 | p0'' < p1''	= p0'
		   | otherwise		= p1'
		p1 | p0'' < p1''	= p1'
		   | otherwise		= p0'

{-
intersectRaySphere :: Ray -> Sphere -> Maybe (Point, Point)
intersectRaySphere (Ray origin direction) (Sphere center radius)
	| tm < 0 && oc2 > r2	= Nothing
	| dt2 < 0		= Nothing
	| otherwise		= Just (p0, p1)
	where	r2 = radius ^ 2
		oc = center `vminus` origin
		oc2 = oc `vdot` oc
		tm = direction `vdot` oc
		lm2 = oc2 - tm^2
		dt2 = r2 - lm2
		t0 = tm - sqrt dt2	
		t1 = tm - sqrt dt2
		p0' = origin `vplus` direction `vsmultiply` t0
		p1' = origin `vplus` direction `vsmultiply` t1
		p0'' = vlength $ p0' `vminus` origin
		p1'' = vlength $ p1' `vminus` origin
		p0 | p0'' < p1''	= p0'
		   | otherwise		= p1'
		p1 | p0'' < p1''	= p1'
		   | otherwise		= p0'
-}

instance NFData (Vector)
	where	rnf (Vector x y z) = x `seq` y `seq` z `seq` ()

instance NFData (Color)
	where	rnf (Color r g b) = r `seq` g `seq` b `seq` ()

renderScene :: [Object] -> [Light] -> [Color] -- returns a list of rows
renderScene objects lights = (concat . parMap rnf renderRow) [0 .. heightI - 1]
			where	renderRow y = map (\x -> renderPixel (x, y)) [0 .. widthI - 1]
					where	renderPixel :: (Int, Int) -> Color
						--renderPixel (x, y) = computeRay (primaryRayThroughPixel x y)
						renderPixel (x, y) = csum $ map (`csdivide` fromIntegral (antiAliasing^2)) $ [computeRay (primaryRayThroughPixelAA x y aaX aaY antiAliasing) reflectionLevel | aaX <- [1..antiAliasing], aaY <- [1..antiAliasing]]
							where	computeRay :: Ray -> ReflectionLevel -> Color
								computeRay ray rl
									| intersectingObjects == []	= backgroundColor
									| rl > 0			= (pointColor `csmultiply` (1 - (reflectiveness material))) `cplus` (reflectionColor `csmultiply` (reflectiveness material))
									| otherwise			= pointColor
									where	pointColor = (oColor closestIntersectingObject) `cmultiply` (ambientComponent `cplus` csum (map (fst . lightToShading) lights)) `cplus` (csum (map (snd . lightToShading) lights))
										reflectionColor = computeRay (Ray closestIntersectionPoint (reflect (direction ray) (oNormal closestIntersectingObject closestIntersectionPoint))) (rl - 1)
										ambientComponent = (ambientLight) `csmultiply` (ambientReflectiveness material)
										lightToShading :: Light -> (Color, Color)
										lightToShading light
											| none (\a -> intersection (oObject a) (Ray closestIntersectionPoint (vnormalize $ (position light) `vminus` closestIntersectionPoint)) /= Nothing) (allExcept closestIntersectingObject objects) = (diffuseComponent, specularComponent)
											| otherwise = (shadowColor, shadowColor)
											where	diffuseComponent = (cmap (clamp 0 1)) $ (diffuse light) `csmultiply` ((diffuseReflectiveness material)*(l `vdot` n))
												specularComponent = (cmap (clamp 0 1)) $ (specular light) `csmultiply` ((specularReflectiveness material)*(max 0 (r `vdot` v))**(shininess material))
												l = vnormalize $ closestIntersectionPoint `vminus` (position light)
												n = vnormalize $ oNormal closestIntersectingObject closestIntersectionPoint
												r = reflect l n
												v = vnormalize $ viewer `vminus` closestIntersectionPoint
										material = oMaterial closestIntersectingObject
										closestIntersectingObject = [x | (x, y) <- zip intersectingObjects intersectingObjectsDistances, y == minimum intersectingObjectsDistances] !! 0 -- should be only one unless two objects share the point, in which case it doesn't matter which one we pick anyway
											where	intersectingObjectsDistances = map (\o -> vlength (distanceTo (fst (just (intersection (oObject o) ray))) (oObject o))) intersectingObjects
										distanceTo :: Point -> Object' -> Vector
										distanceTo p (Sphere' s) = (center s) `vminus` p
										distanceTo p (Triangle' t) = (p1 t) `vminus` p -- FIXME: temporary approximation TODO
										closestIntersectionPoint = (fst (just (intersection (oObject closestIntersectingObject) ray)))
										-- The use of "first" instead of "filter" on the following line is a hack/optimisation for
										-- flat L-systems. It should be changed back to "filter" if anything else is being
										-- rendered.
										intersectingObjects = filter (intersects) objects
											where	intersects :: Object -> Bool
												intersects (Object obj mat col) = intersection obj ray /= Nothing

reflect :: Vector -> Vector -> Vector
reflect l n = l `vminus` ((n `vsmultiply` (n `vdot` l)) `vsmultiply` 2)

intersection :: Object' -> Ray -> Maybe (Point, Point)
intersection (Sphere' s) r = intersectRaySphere r s
intersection (Triangle' t) r = intersectRayTriangle r t

renderedSceneToPPM :: [Color] -> String
renderedSceneToPPM c =	   "P3" ++ "\n"
			++ "#a" ++ "\n"
			++ show widthI ++ " " ++ show heightI ++ "\n"
			++ "255" ++ "\n"
			++ unlines (reformat c)
	where	reformat :: [Color] -> [String]
		reformat = map reformat'
			where	reformat' :: Color -> String
				reformat' (Color x y z) = show r ++ " " ++ show g ++ " " ++ show b
					where	r, g, b :: Int
						r = clamp 0 255 (truncate (x * 255))
						g = clamp 0 255 (truncate (y * 255))
						b = clamp 0 255 (truncate (z * 255))

scene :: [Object]
scene = [ (Object (Sphere' (Sphere (Vector (-3) 0 20) 5)) (Material 1 1 1 64 0.5 0 0) (Color 1 0 0)), (Object (Sphere' (Sphere (Vector (-1.25) (-1.25) 5) 1)) (Material 1 1 1 64 0.5 0 0) (Color 0.1 1 0)), (Object (Sphere' (Sphere (Vector 1 1 5) 0.75)) (Material 1 1 1 64 0.5 0 0) (Color 0 0 1)), (Object (Sphere' (Sphere (Vector 2 (-1) 4) 1)) (Material 1 1 1 64 0.5 0 0) (Color 1 1 1)) ]

lights :: [Light]
--lights = [(Light (Vector 5 2 5) (Color 1 1 1) (Color 1 1 1))]
lights = [(Light (Vector (-2) 0 3) (Color 1 1 1) (Color 1 1 1))]--, (Light (Vector 15 15 20) (Color 1 1 1) (Color 1 1 1))]

-----------------------------------------
-- Custom type definitions - L-systems --
-----------------------------------------

-- Some bits here shamelessly copied from the tutorial
-- Others largely inspired from it

-- color, direction faced, position
type Turtle = (Color, Vector, Point)

-- This is an intermediate type before converting to Objects
-- color, start, end (of line segment)
type DrawnSegment = (Color, Point, Point)

data Axis = X | Y | Z

data Command =	  Go Distance
		| Turn Axis Angle
		| Sit
		| Command :#: Command
		| Branch Command
		| GrabPen Color

degreesToRadians :: (Floating a) => a -> a
degreesToRadians = (*) (pi / 180)

execute :: Command -> [DrawnSegment]
execute c = segments
	where	(segments, turtle) = f c ((Color 1 0 0), (Vector 0 1 0), (Vector 120 (-150) 345))
		f :: Command -> Turtle -> ([DrawnSegment], Turtle)
		f (a :#: b) turtle = (asegs ++ bsegs, bturtle)
			where	(asegs, aturtle) = f a turtle
				(bsegs, bturtle) = f b aturtle
		f (Branch a) turtle = f a turtle
		f (Go dst) (color, direction, position) = ([(color, position, newPosition)], (color, direction, newPosition))
			where newPosition = position `vplus` (direction `vsmultiply` dst)
		f (Turn Z b) (color, (Vector x y z), position) = ([], (color, (Vector (x * cos a - y * sin a) (x * sin a + y * cos a) z), position))
			where a = degreesToRadians b
		f (Turn Y b) (color, (Vector x y z), position) = ([], (color, (Vector (z * cos a - x * sin a) y (z * sin a + x * cos a)), position))
			where a = degreesToRadians b
		f (Turn X b) (color, (Vector x y z), position) = ([], (color, (Vector x (y * cos a - z * sin a) (y * sin a + z * cos a)), position))
			where a = degreesToRadians b
		f (GrabPen c) (color, direction, position) = ([], (c, direction, position))
		f (Sit) turtle = ([], turtle)

drawnSegmentsToSpheres :: [DrawnSegment] -> [Object]
drawnSegmentsToSpheres [] = []
drawnSegmentsToSpheres ((color, start, finish):xs) = (map positionToObject splitUpVector) ++ drawnSegmentsToSpheres xs
	where	splitUpVector = map (\a -> start `vplus` (base `vsmultiply` a)) [1 .. factor]
		vector = finish `vminus` start
		base = vector `vsdivide` factor
		factor = 3

		positionToObject pos = (Object (Sphere' (Sphere pos radius)) (Material 1 1 1 64 0.5 0 0) color)
		radius = vlength base / 2.5

arrowhead x = f x
	where	f 0 = GrabPen (Color 1 0 0) :#: Go 10
		f (x+1) = g x :#: p :#: f x :#: p :#: g x
		g 0 = GrabPen (Color 0 0 1) :#: Go 10
		g (x+1) = f x :#: n :#: g x :#: n :#: f x
		p = Turn Z 60
		n = Turn Z (-60)

-- To see a simple scene with four spheres, uncomment the commented line and vice versa.

main :: IO ()
main = putStr $ renderedSceneToPPM $ renderScene (drawnSegmentsToSpheres $ execute (arrowhead 5)) lights
--main = putStr $ renderedSceneToPPM $ renderScene scene lights
